/*************************************************************************
 *                                                                       *
 * Tokamak Physics Engine, Copyright (C) 2002-2007 David Lam.            *
 * All rights reserved.  Email: david@tokamakphysics.com                 *
 *                       Web: www.tokamakphysics.com                     *
 *                                                                       *
 * This library is distributed in the hope that it will be useful,       *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the files    *
 * LICENSE.TXT for more details.                                         *
 *                                                                       *
 *************************************************************************/

#include "math/ne_type.h"
#include "math/ne_debug.h"
#include "tokamak.h"
#include "containers.h"
#include "scenery.h"
#include "collision.h"
#include "constraint.h"
#include "rigidbody.h"
#include "scenery.h"

#ifdef _WIN32
#include <windows.h>
#endif

#include "stack.h"
#include "simulator.h"
#include "perfwin32.h"

nePerformanceData * nePerformanceData::Create(){
	return new nePerformanceData;
}

LARGE_INTEGER perfFreq;
LARGE_INTEGER counter;

/****************************************************************************
*
*	nePerformanceData::Start
*
****************************************************************************/ 
void DunselFunction() { 
  return; 
}

void nePerformanceData::Init(){
  Reset();
  void (*pFunc)() = DunselFunction;
  // Assume the worst
  if ( QueryPerformanceFrequency(&perfFreq) ) {
    // We can use hires timer, determine overhead
    overheadTicks = 200;
    for ( int i=0; i < 20; i++ ) {
      LARGE_INTEGER b,e;
      int Ticks;
      QueryPerformanceCounter(&b);
      (*pFunc)();
      QueryPerformanceCounter(&e);
      Ticks = e.LowPart - b.LowPart;
      if ( Ticks >= 0 && Ticks < overheadTicks ){
         overheadTicks = Ticks;
      }
    }
    // See if Freq fits in 32 bits; if not lose some precision
    perfFreqAdjust = 0;
    int High32 = perfFreq.HighPart;
    while ( High32 ) {
      High32 >>= 1;
      perfFreqAdjust++;
    }
  }
}

void nePerformanceData::Start(){
	Reset();
	QueryPerformanceCounter(&counter);
}

f32 nePerformanceData::GetCount(){
	LARGE_INTEGER tStart, tStop;
	LARGE_INTEGER Freq = perfFreq;
	int Oht = overheadTicks;
	int ReduceMag = 0;
	tStart = counter;
	QueryPerformanceCounter(&tStop);
	counter = tStop;
	unsigned int High32 = tStop.HighPart - counter.HighPart;
	while ( High32 ){
		High32 >>= 1;
		ReduceMag++;
	}
	if ( perfFreqAdjust || ReduceMag ){
    if ( perfFreqAdjust > ReduceMag ){
			ReduceMag = perfFreqAdjust;
    }
    tStart.QuadPart = Int64ShrlMod32(tStart.QuadPart, ReduceMag);
		tStop.QuadPart = Int64ShrlMod32(tStop.QuadPart, ReduceMag);
		Freq.QuadPart = Int64ShrlMod32(Freq.QuadPart, ReduceMag);
		Oht >>= ReduceMag;
	}
	double time;
	// Reduced numbers to 32 bits, now can do the math
  if ( Freq.LowPart == 0 ){
		time = 0.0;
  }
	else
  {
		time = ((double)(tStop.LowPart - tStart.LowPart - Oht))/Freq.LowPart;
  }
	return (f32)time;
}

void nePerformanceData::UpdateDynamic(){
	dynamic += GetCount();
}
void nePerformanceData::UpdatePosition(){
	position += GetCount();
}
void nePerformanceData::UpdateConstrain1(){
	constrain_1 += GetCount();
}
void nePerformanceData::UpdateConstrain2(){
	constrain_2 += GetCount();
}
void nePerformanceData::UpdateCD(){
	cd += GetCount();
}
void nePerformanceData::UpdateCDCulling(){
	cdCulling += GetCount();
}
void nePerformanceData::UpdateTerrain(){
	terrain += GetCount();
}
void nePerformanceData::UpdateControllerCallback(){
	controllerCallback += GetCount();
}
void nePerformanceData::UpdateTerrainCulling(){
	terrainCulling += GetCount();
}
